/*
 * File         : BlockRangeRep.C
 * Author       : Bryan Carpenter
 * Revision     : $Revision: 1.16 $
 * Updated      : $Date: 2000/08/25 19:00:24 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

// Operations on block distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"

inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRangeRep': Block-distributed template range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* BlockRangeRep :: shell() const {
  return new CollapsedRangeRep(blockSize) ;
}

void BlockRangeRep :: location(Location* loc, const int glb) const {

  loc->sub = glb % blockSize ;
  loc->crd = glb / blockSize ;
  loc->dim = dimension ;
}

void BlockRangeRep :: block(Block* blk, const int ker_glb) const {

  const int blockStart  = ker_glb * blockSize ;

  blk->sub_stp = 1 ;
  blk->glb_stp = 1 ;

  if(blockStart + blockSize > extent)
    blk->count = extent - blockStart ;
  else
    blk->count = blockSize ;

  blk->sub_bas = 0 ;
  blk->glb_bas = blockStart ;
}

void BlockRangeRep :: block2(Block* blk, const int ker_glb,
                             const int count, const int base) const {
  blk->sub_stp = 1 ;
  blk->glb_stp = 1 ;

  const int blockStart = ker_glb * blockSize ;

  const int blockStartOffset  = blockStart - base ;
  const int blockFinishOffset = blockStartOffset + blockSize - 1 ;

  int i_hi, i_lo ;
  i_lo = max(blockStartOffset, 0) ;
  i_hi = min(blockFinishOffset, count - 1) ;
 
  blk->count = i_hi - i_lo + 1 ;

  blk->sub_bas = base + i_lo - blockStart ;
  blk->glb_bas = i_lo ;
}

void BlockRangeRep :: block3(Block* blk, const int ker_glb,
                             const int count, const int base,
                             const int stride) const {
  blk->sub_stp = stride ;
  blk->glb_stp = 1 ;

  const int blockStart = ker_glb * blockSize ;

  const int blockStartOffset  = blockStart - base ;
  const int blockFinishOffset = blockStartOffset + blockSize - 1 ;

  int i_hi, i_lo ;
  if(stride > 0) {
    i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
    i_hi = min(blockFinishOffset / stride, count - 1) ;
  }
  else {
    i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
    i_hi = min(blockStartOffset / stride, count - 1) ;
  }
 
  blk->count = i_hi - i_lo + 1 ;

  blk->sub_bas = base + stride * i_lo - blockStart ;
  blk->glb_bas = i_lo ;
}

void BlockRangeRep :: crds(int* ker_count,
                           int* ker_bas, int* ker_str) const {

  *ker_count = extent ? ker_hi + 1 : 0 ;
  *ker_bas   = 0 ;
  *ker_str   = 1 ;
}

void BlockRangeRep :: crds2(int* ker_count, int* ker_bas, int* ker_str,
                            const int count, const int base) const {

  const int ker_top = (base + count - 1) / blockSize ;

  *ker_bas = base / blockSize ;

  *ker_count = extent ? ker_top - *ker_bas + 1 : 0 ;
  *ker_str   = 1 ;
}

void BlockRangeRep :: crds3(int* ker_count, int* ker_bas, int* ker_str,
                            const int count, const int base,
                            const int stride) const {

  const int ker_top = (base + stride * (count - 1)) / blockSize ;

  *ker_bas = base / blockSize ;

  if(stride > 0) {
    *ker_count = extent ? ker_top - *ker_bas + 1 : 0 ;
    *ker_str   = 1 ;
  }
  else {
    *ker_count = extent ? *ker_bas - ker_top + 1 : 0 ;
    *ker_str   = -1 ;
  }
}

int BlockRangeRep :: volume() const {return blockSize ; }

int BlockRangeRep :: glb(const int sub, const int ker_glb) const {
  return ker_glb * blockSize + sub ;
}

int BlockRangeRep :: isAligned(RangeRep* x) const {
  if(x == (RangeRep*) this) return 1 ;
  if(x->format() == DIST_BLOCK && x->dim() == dimension) {
    BlockRangeRep* t = (BlockRangeRep*) x->tem() ;
    return t->blockSize == blockSize && t->extent == extent &&
           x->size() == extent && x->str() == 1 ;
  }
  else
    return 0 ;
}

void BlockRangeRep :: show() {
  AdlibPrintf("BlockRangeRep :: show: size = %d ker = %p lev = %d\n",
              extent, kernel, level) ;
}

#ifdef USE_POOL
MEMORY_POOL_DEFN(BlockRangeRep)
#endif // USE_POOL


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedBlockRangeRep', `ExtendedBlockSubRangeRep':        //
//     Block-distributed range of array with ghost areas.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* ExtendedBlockRangeRep :: shell() const {
  return new ExtendedCollapsedRangeRep(blockSize, wlo, whi) ;
}

int ExtendedBlockRangeRep :: loExtension() const {
  return wlo;
}

int ExtendedBlockRangeRep :: hiExtension() const {
  return whi;
}

//int ExtendedBlockRangeRep :: disp(const int sub) const {
//  return sub + wlo ;
//}

int ExtendedBlockRangeRep :: offset(const Location* i) const {
  //return i->sub + wlo ;
  return i->sub ;
}

int ExtendedBlockRangeRep :: disp() const {
  return wlo ;
}

int ExtendedBlockRangeRep :: volume() const {
  return wlo + blockSize + whi ;
}

#ifdef USE_POOL
MEMORY_POOL_DEFN(ExtendedBlockRangeRep)
#endif // USE_POOL


//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
//// classes `PackedBlockRangeRep':                                      //
////      Block-distributed ranges of array with ``packed'' layout.      //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//RangeRep* PackedBlockRangeRep :: shell() const {
//  AdlibError("PackedBlockRangeRep :: shell: "
//             "member undefined on this range") ;
//  return 0 ;
//}
//
//int PackedBlockRangeRep :: disp(const int sub) const {
//  return sub / packing ;
//}
//
//int PackedBlockRangeRep :: step(const int sub_stp) const {
//  return sub_stp / packing ;
//}
//
//int PackedBlockRangeRep :: offset(const Location* i) const {
//  return i->sub / packing ;
//}
//
//int PackedBlockRangeRep :: volume() const {
//  return width ;
//}
//
//#ifdef USE_POOL
//MEMORY_POOL_DEFN(PackedBlockRangeRep)
//#endif // USE_POOL


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockCyclicRangeRep': Block-cyclically distributed ranges.   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void BlockCyclicRangeRep :: location(Location* loc, const int glb) const {
  const int numProcs = dimension.size() ;

  const int ker = glb / blockSize ;

  loc->sub = glb % blockSize ;
  loc->blk = ker / numProcs ;
  loc->crd = ker % numProcs ;
  loc->dim = dimension ;
}

int BlockCyclicRangeRep :: offset(const Location* i) const {
  return i->sub + blockSize * i->blk ;
}

int BlockCyclicRangeRep :: volume() const {
  return blockSize * kernel->volume() ;
}

int BlockCyclicRangeRep :: isAligned(RangeRep* x) const {
  if(x == (RangeRep*) this) return 1 ;
  if(x->format() == DIST_BLOCK_CYCLIC && kernel->isAligned(x->ker())) {
    BlockCyclicRangeRep* t = (BlockCyclicRangeRep*) x->tem() ;
    return t->blockSize == blockSize && t->extent == extent &&
           x->size() == extent && x->str() == 1 ;
  }
  else
    return 0 ;
}


//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
//// class `PackedBlockCyclicRangeRep':                                  //
////      Block-cyclically distributed range of array with ``packed''    //
////      layout.                                                        //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//void PackedBlockCyclicRangeRep :: location(Location* loc,
//                                           const int glb) const {
//  const int numProcs = dimension.size() ;
//
//  const int ker = glb / blockSize ;
//
//  loc->sub = glb % blockSize ;
//  loc->blk = ker / numProcs ;
//  loc->crd = ker % numProcs ;
//  loc->dim = dimension ;
//}
//
//int PackedBlockCyclicRangeRep :: offset(const Location* i) const {
//  return i->sub / packing + width * i->blk ;
//}
//
//int PackedBlockCyclicRangeRep :: volume() const {
//  return width * kernel->volume() ;
//}


// -------------------------------------------------------------
// Things to do
//
//   Extended block-cyclic subclasses.
//
//   `show' methods for subclasses.

